"""
Command Design Pattern (Behavioral Design Pattern)
"""
import uuid
from typing import Dict
from zesty.models.hf_interface import *
from zesty.models.hf_interface import *
from zesty.models.disk_mon import FileSystem


class ZBSActionData:
    mount_path = None
    device = None
    filesystem = None
    fs_type = None
    is_partition = False
    partition_number = None
    LV = None
    VG = None
    lvm_path = None
    chunk_size = None

    def __init__(self, mount_path=None,
                 device=None,
                 filesystem=None,
                 fs_type=None,
                 is_partition=False,
                 partition_number=None,
                 LV=None,
                 VG=None,
                 lvm_path=None,
                 chunk_size=None):
        self.mount_path = mount_path
        self.filesystem = filesystem
        self.fs_type = fs_type
        self.device = device
        self.is_partition = is_partition
        self.partition_number = partition_number
        self.LV = LV
        self.VG = VG
        self.lvm_path = lvm_path
        self.chunk_size = chunk_size

    def serialize(self):
        return self.__dict__

    def set_data(self, json):
        self.mount_path = json.get('mount_path')
        self.filesystem = json.get('filesystem')
        self.fs_type = json.get('fs_type')
        self.device = json.get('device')
        self.is_partition = json.get('is_partition', False)
        self.partition_number = json.get('partition_number', '')
        self.LV = json.get('LV', '')
        self.VG = json.get('VG', '')
        self.lvm_path = json.get('lvm_path', '')
        self.chunk_size = json.get('chunk_size', 0)

        return self


class ZBSAgentReceiver:
    """
    The ZBSAgentReceiver (Receiver class in the Command pattern) contain some important business logic.
    It knows how to perform any kind of action sent by the ZBS Backend.
    ZBSAgent is an abstract class, while the concrete implementations should be per OS
    """

    @abstractmethod
    def do_nothing(self, data: ZBSActionData) -> None:
        raise NotImplementedError(
            "ZBSAgentReceiver 'do_nothing' is abstract, please implement a concrete per OD receiver")

    @abstractmethod
    def extend_fs(self, data: ZBSActionData) -> None:
        raise NotImplementedError(
            "ZBSAgentReceiver 'extend_fs' is abstract, please implement a concrete per OD receiver")

    @abstractmethod
    def add_disk(self, data: ZBSActionData) -> None:
        raise NotImplementedError(
            "ZBSAgentReceiver 'add_disk' is abstract, please implement a concrete per OD receiver")

    @abstractmethod
    def balance_fs(self, data: ZBSActionData) -> None:
        raise NotImplementedError(
            "ZBSAgentReceiver 'balance_fs' is abstract, please implement a concrete per OD receiver")

    @abstractmethod
    def remove_disk(self, data: ZBSActionData) -> None:
        raise NotImplementedError(
            "ZBSAgentReceiver 'remove_disk' is abstract, please implement a concrete per OD receiver")

    @abstractmethod
    def balance_ebs_structure(self, data: ZBSActionData) -> None:
        raise NotImplementedError(
            "ZBSAgentReceiver 'balance_ebs_structure' is abstract, please implement a concrete per OD receiver")


class ZBSAction(IActionHF):
    """
    Base command class
    Delegates the business logic to the receiver
    There are receivers per OS (Linux and Windows for now)
    """
    TYPE_FIELD_NAME = "type"
    DATA_FIELD_NAME = "data"

    __uuid = uuid.uuid4()
    __status: IActionHF.Status = IActionHF.Status.NEW
    __special_instructions: SpecialInstructions

    def __init__(self, receiver: ZBSAgentReceiver = None, data: ZBSActionData = None):
        self.receiver = receiver
        self.data = data

    def set_data(self, data: ZBSActionData):
        self.data = data

    def set_receiver(self, receiver: ZBSAgentReceiver):
        self.receiver = receiver

    def serialize(self):
        result = self.__dict__
        result[ZBSAction.TYPE_FIELD_NAME] = self.get_action_type()
        result[ZBSAction.DATA_FIELD_NAME] = self.data.serialize()

        return result

    # ActionHF interface implementation
    def get_action_id(self) -> str:
        return str(self.__uuid)

    def get_action_type(self) -> str:
        return str(type(self).__name__)

    def get_status(self) -> IActionHF.Status:
        return self.__status

    def set_status(self, status: IActionHF.Status):
        self.__status = status

    def get_special_instructions(self) -> SpecialInstructions:
        return self.__special_instructions

    def set_special_instructions(self, special_instructions: SpecialInstructions):
        self.__special_instructions = special_instructions

    @staticmethod
    def deserialize_type(json):
        return json[ZBSAction.TYPE_FIELD_NAME]

    @staticmethod
    def deserialize_data(json):
        return ZBSActionData().set_data(json[ZBSAction.DATA_FIELD_NAME])

    @abstractmethod
    def execute(self):
        raise NotImplementedError("BaseAction is abstract, please implement a concrete action")


class FileSystemHF(IFileSystemHF):
    fs_id: str
    devices: Dict[str, FileSystem.BlockDevice]
    existing_actions: Dict[str, ZBSAction]

    def __init__(self, fs_id: str, devices: Dict[str, FileSystem.BlockDevice], existing_actions: Dict[str, ZBSAction]):
        self.fs_id = fs_id
        self.devices = devices
        self.existing_actions = existing_actions

    def get_fs_id(self) -> str:
        return self.fs_id

    def get_devices(self) -> Dict[str, FileSystem.BlockDevice]:
        return self.devices

    def get_existing_actions(self) -> Dict[str, ZBSAction]:
        return self.existing_actions


class DoNothingAction(ZBSAction):
    """
    Do nothing action
    """

    def execute(self):
        print("Do nothing")

    class Factory:
        def create(self): return DoNothingAction()


class ExtendFileSystemAction(ZBSAction):
    """
    Extend File System Action.
    """

    def execute(self):
        try:
            self.receiver.extend_fs(self.data)
        except AttributeError as ex:
            print("Failed to execute command '{}': error is '{}'".format(self.get_action_type(), ex))

    class Factory:
        def create(self): return ExtendFileSystemAction()


class AddDiskAction(ZBSAction):
    """
    Add Disk Action.
    """

    def execute(self):
        try:
            self.receiver.add_disk(self.data)
        except AttributeError as ex:
            print("Failed to execute command '{}': error is '{}'".format(self.get_action_type(), ex))

    class Factory:
        def create(self): return AddDiskAction()


class RemoveDiskAction(ZBSAction):
    """
    Remove Disk Action.
    """

    def execute(self):
        try:
            self.receiver.remove_disk(self.data)
        except AttributeError as ex:
            print("Failed to execute command '{}': error is '{}'".format(self.get_action_type(), ex))

    class Factory:
        def create(self): return RemoveDiskAction()


class BalanceFileSystemAction(ZBSAction):
    """
    Balance File System Action.
    """

    def execute(self):
        try:
            self.receiver.balance_fs(self.data)
        except AttributeError as ex:
            print("Failed to execute command '{}': error is '{}'".format(self.get_action_type(), ex))

    class Factory:
        def create(self): return BalanceFileSystemAction()


class BalanceEBSStructureAction(ZBSAction):
    """
    Balance EBS structure Action.
    """

    def execute(self):
        try:
            self.receiver.extend_fs(self.data)
            self.receiver.remove_disk(self.data)
        except AttributeError as ex:
            print("Failed to execute command '{}': error is '{}'".format(self.get_action_type(), ex))

    class Factory:
        def create(self): return BalanceEBSStructureAction()


class ZBSActionFactory:
    actions = {}

    def create_action(action_type):
        if action_type not in ZBSActionFactory.actions:
            ZBSActionFactory.actions[action_type] = eval(action_type + '.Factory()')
        return ZBSActionFactory.actions[action_type].create()

    create_action = staticmethod(create_action)
